/*
 * Linux console program which display in real time users currently logged in.
 * You can see user's processes and type of login (if it was via telnet,
 * ssh, or something else). 
 * You have to have read access to utmp, wtmp and /proc
 * 
 * compile:
 *	gcc whowatch.c -o whowatch -lncurses
 * 
 * You can also add options:
 * 	-DNOCOLORS	don't use colors
 * 	-DNOAUTO	don't refresh processes information. You can
 *			refresh this information manually by  pressing 'w' key 
 *			while program is running
 *
 * author: Michal Suszycki	mike@wizard.ae.krakow.pl
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */
 
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <utmp.h>
#include <string.h>
#include <signal.h>
#include <curses.h>

#define UTMP 	"/var/run/utmp"
#define WTMP 	"/var/log/wtmp"
#define SSH 	"(sshd)"
#define TELNET 	"(in.telnetd)"

/* uncomment this if you want to disable automatic refresh of processes info.
   (to safe CPU time)
   In this case it can be done manually by pressing 'w' key.
*/
//#define NOAUTO 

/* uncomment this if you don't want colors */
//#define NOCOLOR 

/* number of lines reserved for header info */
#define TOP 2

/* number of seconds before another attempt of reading wtmp*/
#define TIMEOUT	2

/* number of bottom line which we don't use for printing users.
   (first line is number 0)
   LINES is defined in <curses.h> and will be filled in by initscr 
   with the  size of the screen.   */

#define BOTTOM LINES-1

/*--------- end of user's configuration section -----------------------*/


#define CMD_COLUMN 52

/* can we use colors? */
int color;

FILE *debug_file;


struct user
{
	char *name;
	char *tty;
	int prot;
	int pid;
	struct user **prev;
	struct user *next;
	int line;
};


int how_many, telnet_users, ssh_users;
struct user *begin ;

void print_user(struct user *pentry,struct utmp *putmp);
struct utmp *get_utmp_user();
char *get_w();
int last_line;
int fdwtmp;

void allocate_error(){
	fprintf(stderr,"Cannot allocate memory.\n");
	exit (-1);
}

// Find parent pid (in /proc/pid/stat)

int get_ppid(int pid)
{
        static int ppid;
        char buff[256];
        FILE *f;
        sprintf(buff,"/proc/%d/stat",pid);
        if (!(f =  fopen(buff,"rt")))
		return -1;
        fscanf(f,"%*d %*s %*c %d",&ppid);
        fclose(f);
        return ppid;
}
														
char *get_name(int pid)
{
        static char buff[256];
        char name[64];
        FILE *f;
        int i = 0;
        sprintf(buff,"/proc/%d/stat",pid);
        if (!(f = fopen(buff,"rt")))
               	return "can't read";

        fscanf(f,"%*d %s",name);
        while ((buff[i] = name[i])) i++;
        buff[i] = '\0';
	fclose(f);
	return buff;
}

char *get_cmdline(int pid)
{
        static char buff[512];
        FILE *f;
        int i = 0;
        sprintf(buff,"/proc/%d/cmdline",pid);
        if (!(f = fopen(buff,"rt")))
               	return "-";

	while (fread(buff+i,1,1,f) == 1){
		if (buff[i] == '\0') buff[i] = ' ';
		if (i == 512 - 2) break;
		i++;
	}
        buff[i] = '\0';
	fclose(f);
	return buff;
}
															
void update_info()
{
	char buf[128];

	if (color) attrset(COLOR_PAIR(1));
	sprintf(buf,"Users: %d (telnet %d , ssh %d , other %d )  ",
	  how_many, telnet_users, ssh_users, how_many - telnet_users - ssh_users);
	move(0,0);
	addstr(buf);
	move(last_line,0);
	if (color) attrset(A_BOLD);
	refresh();
}

void read_utmp()
{
	int fd, ppid, lines;
	static struct utmp entry;
	struct user *tmp, **current = &begin;
	char *count;
	
	if ((fd = open(UTMP,O_RDONLY)) == -1){
		fprintf(stderr,"Cannot open " UTMP "\n");
		exit (-1);
	}
	
	tmp = NULL;
	move(TOP,0);
	lines = TOP;
	last_line = TOP;
	
	while(read(fd, &entry,sizeof entry) > 0){
		
		if (entry.ut_type != 7) continue;
		
		if (!(tmp = malloc(sizeof (struct user)) ))
			allocate_error();
		
		if (!(tmp->name = malloc(strlen(entry.ut_user) + 1)))
			allocate_error();
		if (!(tmp->tty = malloc(strlen(entry.ut_line) + 1)))
			allocate_error();	
		strncpy(tmp->name,entry.ut_user,8);
		strcpy(tmp->tty,entry.ut_line);
		
		tmp->pid = entry.ut_pid;
		
		how_many++;
		ppid = get_ppid(tmp->pid);
		count = get_name(ppid);
		
		if (!strcmp(TELNET,count)){
			telnet_users++;
			tmp->prot = 23;
		}
		else if (!strcmp(SSH,count)){
			tmp->prot = 22;
			ssh_users++;
		} else 
			tmp->prot = 0;
		
		
		if (lines >= BOTTOM)
			tmp->line = -1;
		else {
			tmp->line = lines;
			print_user(tmp,&entry);
		}
		lines++;
		
		*current = tmp;
		tmp->prev = current;
		current = &tmp->next;

#ifdef DEBUG
		fprintf(debug_file,"%-8s %-5s (%-12d <- %-12d)%d\n",
			tmp->name, tmp->tty, *tmp->prev,tmp,tmp->line);
#endif
		
	}
	*current = NULL;
	fflush(debug_file);
	close(fd);
	return;
}



struct user* new_user(struct utmp *newone)
{
	struct user *i = begin,*tmp;
	char *count;
	int ppid;
	
	while(i->next) i=i->next;
	if (!(tmp = malloc(sizeof (struct user)) ))
		allocate_error();

	if (!(tmp->name = malloc(UT_NAMESIZE + 1)))
		allocate_error();
	tmp->name[UT_NAMESIZE] = '\0';
	strncpy(tmp->name,newone->ut_user,UT_NAMESIZE);

	if (!(tmp->tty = malloc(UT_LINESIZE + 1)))
		allocate_error();
	tmp->tty[UT_LINESIZE] = '\0';	
	strncpy(tmp->tty,newone->ut_line,UT_LINESIZE);
	
	tmp->pid = newone->ut_pid;
	
	tmp->next = NULL;
	tmp->prev = &i->next;
	i->next = tmp;	
	
	if (last_line >= BOTTOM) 
		tmp->line = -1;
	else
		tmp->line = last_line;
	
	if ((ppid = get_ppid(newone->ut_pid)) == -1)
		count = "can't access";
	else
		count = get_name(ppid);
		
	if (!strcmp(SSH,count)){
		tmp->prot = 22;
		ssh_users++;
	}
	else if (!strcmp(TELNET,count)){
		tmp->prot = 23;
		telnet_users++;
	}
	else 
		tmp->prot = 0;
			
	how_many++;
	return tmp;
}
	
void update_line_nr(int line)
{
	struct user *tmp = begin;
	while(tmp){
		if (tmp->line > line && tmp->line != -1) tmp->line--;
		tmp = tmp->next;
	}
}

void print_user(struct user *pentry, struct utmp *putmp)
{
	char buff[256];
	int ppid;
	char *count;
	
	if (last_line >= BOTTOM) return;
	
	if ((ppid = get_ppid(pentry->pid)) == -1)
		count = "can't access";
	else 
		count = get_name(ppid);
		
	sprintf(buff,"%-15s%-10s%-7s%-20s%-20s\n",count,pentry->name,
			pentry->tty, putmp?putmp->ut_host:"unknown",get_w(pentry->pid));
	
	if (color) attrset(A_BOLD);
	addnstr(buff,COLS);
	last_line++;
	refresh();
}

struct utmp *get_utmp_user(char *tty)
{
	int fd;
	static struct utmp entry;
	
	if ((fd = open(UTMP,O_RDONLY)) == -1){
		fprintf(stderr,"Cannot open " UTMP "\n");
		exit (-1);
	}
	
	while(read(fd, &entry,sizeof entry) > 0){
		if (entry.ut_type != 7) continue;
		if (!strcmp(tty,entry.ut_line)) {
			close(fd);
			return &entry;
		}
	}
	close(fd);
	return NULL;
}	

void delete_user(struct user *p)
{	
	struct user *tmp = begin;
	
	if (p->prot == 23) 
		telnet_users--;
	if (p->prot == 22)
		ssh_users--; 
	how_many--;

	if (p->line == -1) return;
	
	move(p->line,0);
	deleteln();
	update_line_nr(p->line);
	last_line--;
	move(last_line,0);	
	
	while (tmp){
		if (tmp->line == -1){
			tmp->line = last_line;
			print_user(tmp,get_utmp_user(tmp->tty));
			break;
		}
		tmp = tmp->next;
	}
	
	return;
}

		 	
void show_u()
{
	struct user *f = begin ;
	fprintf(debug_file,"SIGSEGV caught\n");
	fprintf(debug_file,"begin = %d\n",begin);
	while (f){
		fprintf(debug_file,"adress %d\tprev %-11d\tnext %-10d\t(%s,%s) %d\n",
			f,*f->prev,f->next,f->name,f->tty,f->line);
		f = f->next;
	}
	fflush(debug_file);
	endwin();
	fprintf(stderr,"SIGSEGV caught\n");
	kill(getpid(),6);
	exit (0);
}

void cleanup()
{
	struct user *i = begin, *tmp;
	while (i){
		tmp = i->next;
		free(i->name);
		free(i->tty);
		free(i);
		i = tmp;
	}
	begin = NULL;
	close(fdwtmp);
}

void restart()
{
	how_many = 0;
	ssh_users = 0;
	telnet_users = 0;

	read_utmp();
	if ((fdwtmp = open(WTMP,O_RDONLY)) == -1){
		fprintf(stderr,"Cannot open " WTMP "\n");
		exit (-1);
	}
	
	lseek(fdwtmp,0,SEEK_END);
}

void curses_init()
{
	initscr();
	start_color();
	if (COLS < 80 || LINES < 10){
		 fprintf(stderr,"Terminal too small. I expect min. 80 columns and 10 rows.\n");
		 endwin();
		 exit(0);
	}
	if (has_colors() == TRUE && can_change_color() == TRUE){
		color = 1;
		init_pair(1,COLOR_CYAN,COLOR_BLACK);
		init_pair(2,COLOR_GREEN,COLOR_BLACK);
	}
#ifdef NOCOLORS
	color = 0;
#endif
	cbreak();
	nodelay(stdscr,TRUE);
	noecho();
}

void clear_field(int p)
{
	int i, y, x;
	getyx(stdscr,y,x);
	
	for (i = 0 ; i < COLS - p; i++) addch(' ');
	
	move(y,x);
}
	
char *get_w(int pid)
{
	static char buf[256];
	FILE *f;
	static int tpgid;
	
	sprintf(buf,"/proc/%d/stat",pid);
	if (!(f = fopen(buf,"r"))) return "-";
	fscanf(f,"%*d %*s %*c %*d %*d %*d %*d %d",&tpgid);
	fclose(f);
	return get_cmdline(tpgid);
}


void show_cmd()
{
	struct user *tmp = begin;

	while(tmp){
		if (tmp->line == -1) {
			tmp = tmp->next;
			continue;
		}
		move(tmp->line,CMD_COLUMN);
		clear_field(CMD_COLUMN);
		addnstr(get_w(tmp->pid),COLS - CMD_COLUMN);
		tmp = tmp->next;
	}
	move(last_line,0);
	refresh();
}
		
		
void main()
{
	struct user *f, *tmp;
	struct utmp entry;
	
	curses_init();

#ifdef DEBUG	
	signal(SIGINT,(void (*)())show_u);	
	signal(SIGSEGV,(void (*)())show_u);
	
	if (!(debug_file = fopen("./wlogs","a"))){
		perror("fopen");
		exit (-1);
	}

#endif	
	read_utmp();
	update_info();

#ifdef DEBUG
	fprintf(debug_file,"Started. begin = %d\n",begin);
#endif
	if ((fdwtmp = open(WTMP,O_RDONLY)) == -1){
		fprintf(stderr,"Cannot open " WTMP "\n");
		exit (-1);
	}
	
	lseek(fdwtmp,0,SEEK_END);
	while(1){
		while(!read(fdwtmp, &entry,sizeof entry)) {
			int c;
			c = getch();			
#ifdef NOAUTO
			
			if (c != ERR && c == 'w'){
				show_cmd();
			}
#else
			show_cmd();
#endif
			if (c != ERR && c == 'r'){
				cleanup();
				restart();
			}
			sleep(TIMEOUT);
		}
		/* type 7 is USER_PROCESS (user just logged in)*/
		if (entry.ut_type == 7){
			tmp = new_user(&entry);
			print_user(tmp,&entry);
			update_info();
#ifdef DEBUG
			fprintf(debug_file,"IN: %-9s%-5s, %-8s %-5s (%-12d<- %-12d) %d\n",
				entry.ut_user,entry.ut_line,tmp->name,tmp->tty,*tmp->prev,tmp,tmp->line);
			fflush(debug_file);
#endif
			continue;
		}

		if (entry.ut_type != 8) continue;
		
	/* type 8 is DEAD_PROCESS (user just logged out) */
		f = begin;
		while(f){
			if (strncmp(f->tty,entry.ut_line,UT_LINESIZE)) {
				f = f->next;
				continue;
			}
#ifdef DEBUG
			fprintf(debug_file,"OUT %-9s%-5s, %-9s %-5s (%-12d<- %-12d)%d\n",
				entry.ut_user,entry.ut_line,f->name,f->tty,*f->prev,f,f->line);
			fflush(debug_file);
#endif			
			delete_user(f);
			update_info();

			tmp = f->next;
			*(f->prev) = tmp;
			if (tmp) tmp->prev = f->prev;

			free(f->name);
			free(f->tty);
			free(f);
			break;
		}
	}
}

